# Copyright (c) Acconeer AB, 2022-2023
# All rights reserved

from __future__ import annotations

import importlib.resources
import logging
from enum import Enum
from functools import partial
from typing import Any, Optional

import qtawesome as qta

from PySide6 import QtCore, QtGui
from PySide6.QtSvgWidgets import QSvgWidget
from PySide6.QtWidgets import (
    QButtonGroup,
    QFrame,
    QGraphicsOpacityEffect,
    QGridLayout,
    QGroupBox,
    QHBoxLayout,
    QLabel,
    QPushButton,
    QSizePolicy,
    QVBoxLayout,
    QWidget,
)

from acconeer.exptool.app import resources
from acconeer.exptool.app.new._enums import PluginFamily, PluginGeneration, PluginState
from acconeer.exptool.app.new.app_model import AppModel, PluginPresetSpec, PluginSpec
from acconeer.exptool.app.new.pluginbase import PlotPluginBase, PluginSpecBase
from acconeer.exptool.app.new.ui.components.group_box import GroupBox
from acconeer.exptool.app.new.ui.icons import ARROW_LEFT_BOLD, EXTERNAL_LINK, TEXT_GREY
from acconeer.exptool.app.new.ui.utils import (
    HorizontalSeparator,
    LayoutWrapper,
    ScrollAreaDecorator,
    TopAlignDecorator,
)


log = logging.getLogger(__name__)


class PluginSelectionButton(QPushButton):
    plugin: PluginSpec

    def __init__(self, plugin: PluginSpecBase, parent: QWidget) -> None:
        super().__init__(parent)

        self.plugin = plugin

        self.setText(plugin.title)
        self.setStyleSheet("text-align: left; font-weight: bold;")
        self.setCheckable(True)


class PluginSelectionButtonGroup(QButtonGroup):
    def __init__(self, parent: QWidget) -> None:
        super().__init__(parent)

        self.setExclusive(True)

    def checkedButton(self) -> PluginSelectionButton:
        button = super().checkedButton()
        assert isinstance(button, PluginSelectionButton)
        return button

    def buttons(self) -> list[PluginSelectionButton]:  # type: ignore[override]
        buttons = super().buttons()
        assert isinstance(buttons, list)
        assert all(isinstance(e, PluginSelectionButton) for e in buttons)
        return buttons  # type: ignore[return-value]


class PluginPresetButton(QPushButton):
    def __init__(self, plugin_preset: PluginPresetSpec, default: bool, parent: QWidget) -> None:
        super().__init__(parent)

        if default:
            self.setText(f"{plugin_preset.name} (default)")
        else:
            self.setText(plugin_preset.name)
        if plugin_preset.description is not None:
            self.setToolTip(plugin_preset.description)
        self.setStyleSheet("text-align: left; font-weight: bold;")


class PluginPresetPlaceholder(QWidget):
    def __init__(self, app_model: AppModel, parent: Optional[QWidget] = None) -> None:
        super().__init__(parent)

        self.app_model = app_model
        self.app_model.sig_load_plugin.connect(self._on_load_plugin)
        self.app_model.sig_notify.connect(self._on_app_model_update)

        self.app_model = app_model
        self.setLayout(QVBoxLayout(self))
        self.layout().setContentsMargins(0, 0, 0, 0)
        self.layout().setSpacing(11)
        self.group_box = GroupBox.vertical("Preset Configurations", parent=self)
        self.layout().addWidget(self.group_box)
        self.preset_buttons_widget: Optional[QWidget] = None

    def _clear(self) -> None:
        if self.preset_buttons_widget is not None:
            self.preset_buttons_widget.deleteLater()
            self.preset_buttons_widget = None

    def _on_preset_click(self, preset_id: Enum) -> None:
        self.app_model.set_plugin_preset(preset_id)

    def _on_load_plugin(self, plugin_spec: Optional[PluginSpecBase]) -> None:
        self._clear()
        if plugin_spec is not None and plugin_spec.presets is not None:
            self.preset_buttons_widget = QWidget(self.group_box)
            self.preset_buttons_widget.setLayout(QVBoxLayout(self.preset_buttons_widget))
            self.preset_buttons_widget.layout().setContentsMargins(0, 0, 0, 0)
            self.group_box.layout().addWidget(self.preset_buttons_widget)

            for preset in plugin_spec.presets:
                default_preset = (
                    len(plugin_spec.presets) > 1
                    and preset.preset_id == plugin_spec.default_preset_id
                    and preset.name.lower() != "default"
                )
                button = PluginPresetButton(preset, default_preset, self)
                self.preset_buttons_widget.layout().addWidget(button)
                button.clicked.connect(partial(self._on_preset_click, preset.preset_id))

    def _on_app_model_update(self, app_model: AppModel) -> None:
        self.setEnabled(app_model.plugin_state == PluginState.LOADED_IDLE)


class PluginSelection(QWidget):
    def __init__(
        self, app_model: AppModel, plugins: list[PluginSpecBase], parent: QWidget
    ) -> None:
        super().__init__(parent)

        self.app_model = app_model

        app_model.sig_notify.connect(self._on_app_model_update)

        self.setLayout(QVBoxLayout(self))
        self.layout().setContentsMargins(0, 0, 0, 0)
        self.layout().setSpacing(11)

        group_boxes = {}
        for family in PluginFamily:
            group_box = QGroupBox(self)
            group_box.setTitle(family.value)
            group_box.setHidden(True)
            group_box.setLayout(QVBoxLayout(group_box))
            self.layout().addWidget(group_box)
            group_boxes[family] = group_box

        self.button_group = PluginSelectionButtonGroup(self)
        self.button_group.buttonClicked.connect(self._on_load_click)

        for plugin in plugins:
            assert isinstance(plugin, PluginSpecBase)
            group_box = group_boxes[plugin.family]
            group_box.setHidden(False)

            button_layout = QHBoxLayout()
            button_layout.setContentsMargins(0, 0, 0, 0)
            group_box.layout().addWidget(LayoutWrapper(button_layout))

            button = PluginSelectionButton(plugin, group_box)
            self.button_group.addButton(button)
            button_layout.addWidget(button)

            if plugin.description:
                label = QLabel(group_box)
                label.setText(plugin.description)
                label.setWordWrap(True)
                group_box.layout().addWidget(label)

            if plugin.docs_link:
                info_button = QPushButton(text="docs", icon=EXTERNAL_LINK(color=TEXT_GREY))
                info_button.setStyleSheet(f"QPushButton {{color: {TEXT_GREY}}}")
                info_button.setSizePolicy(QSizePolicy.Policy.Fixed, QSizePolicy.Policy.Fixed)
                info_button.clicked.connect(
                    lambda _=False, p=plugin: QtGui.QDesktopServices.openUrl(p.docs_link)
                )
                button_layout.addWidget(info_button)

    def layout(self) -> QVBoxLayout:
        return super().layout()  # type: ignore[return-value]

    def _on_load_click(self) -> None:
        self.layout().addStretch(1)

        plugin = self.button_group.checkedButton().plugin
        self.app_model.load_plugin(plugin)

    def _on_app_model_update(self, app_model: AppModel) -> None:
        plugin: Optional[PluginSpec] = app_model.plugin

        if plugin is None:
            self.button_group.setExclusive(False)

            for button in self.button_group.buttons():
                button.setChecked(False)

            self.button_group.setExclusive(True)
        else:
            buttons = self.button_group.buttons()
            try:
                (button_to_check,) = {b for b in buttons if b.plugin == plugin}
            except ValueError:
                pass
            else:
                button_to_check.setChecked(True)

        self.setEnabled(app_model.plugin_state.is_steady)


class PluginSelectionArea(QWidget):
    def __init__(self, app_model: AppModel, parent: Optional[QWidget] = None) -> None:
        super().__init__(parent)
        self.setLayout(QVBoxLayout(self))
        self.layout().setContentsMargins(0, 0, 0, 0)

        a121_selection = ScrollAreaDecorator(
            TopAlignDecorator(
                PluginSelection(
                    app_model,
                    [
                        plugin
                        for plugin in app_model.plugins
                        if (
                            plugin.generation == PluginGeneration.A121
                            and isinstance(plugin, PluginSpecBase)
                        )
                    ],
                    self,
                )
            )
        )
        a121_selection.setVisible(True)

        app_model.sig_notify.connect(
            lambda am: a121_selection.setVisible(am.plugin_generation == PluginGeneration.A121)
        )

        self.layout().addWidget(a121_selection)
        self.layout().addWidget(HorizontalSeparator())
        self.layout().addWidget(PluginPresetPlaceholder(app_model, self))


class PlotPlaceholder(PlotPluginBase):
    def __init__(self, app_model: AppModel) -> None:
        super().__init__(app_model)

        layout = QVBoxLayout(self)

        layout.addStretch(1)
        layout.addLayout(self._select_module_text())
        layout.addWidget(self._teaser_text())
        layout.addStretch(1)

        self.setLayout(layout)

    def _select_module_text(self) -> QHBoxLayout:
        h_box = QHBoxLayout()
        h_box.addStretch(1)

        icon_widget = qta.IconWidget()
        icon_widget.setIconSize(QtCore.QSize(36, 36))
        icon_widget.setIcon(ARROW_LEFT_BOLD(color="#4d5157"))

        label = QLabel("Select a module to begin", self)
        label.setStyleSheet("font-size: 20px;")
        label.setAlignment(QtCore.Qt.AlignmentFlag.AlignCenter)

        h_box.addWidget(icon_widget)
        h_box.addWidget(label)
        h_box.addStretch(1)

        return h_box

    def _teaser_text(self) -> QLabel:
        teaser_label = QLabel(
            "More detectors and example applications will\nbe released continuously.", self
        )
        teaser_label.setStyleSheet("font-size: 16px;")
        teaser_label.setAlignment(QtCore.Qt.AlignmentFlag.AlignCenter)

        return teaser_label

    def handle_message(self, message: Any) -> None:
        pass

    def draw(self) -> None:
        pass


class PluginPlotArea(QFrame):
    _FPS = 60

    def __init__(self, app_model: AppModel, parent: QWidget) -> None:
        super().__init__(parent)

        self.app_model = app_model

        self.plot_plugin: PlotPluginBase = PlotPlaceholder(app_model)

        self.setObjectName("PluginPlotArea")
        self.setStyleSheet("QFrame#PluginPlotArea {background: #fff; border: 0;}")
        self.setFrameStyle(0)

        self.setLayout(QVBoxLayout(self))
        self.layout().setContentsMargins(0, 0, 0, 0)
        self.layout().setSpacing(0)
        self.layout().addWidget(self.plot_plugin)

        self.startTimer(int(1000 / self._FPS))

        app_model.sig_load_plugin.connect(self._on_app_model_load_plugin)
        if isinstance(app_model.plugin, PluginSpecBase):
            self._on_app_model_load_plugin(app_model.plugin)
        elif app_model.plugin is not None:
            raise RuntimeError(f"{type(app_model.plugin)} is not a PluginSpecBase.")

    def timerEvent(self, event: QtCore.QTimerEvent) -> None:
        self.plot_plugin.draw()

    def _on_app_model_load_plugin(self, plugin: Optional[PluginSpecBase]) -> None:
        log.debug(
            f"{self.__class__.__name__} is going to replace its plot_plugin "
            + f"({self.plot_plugin.__class__.__name__}) in favour of {plugin}"
        )
        self.plot_plugin.deleteLater()

        if plugin is None:
            self.plot_plugin = PlotPlaceholder(self.app_model)
        else:
            self.plot_plugin = plugin.create_plot_plugin(app_model=self.app_model)

        self.layout().addWidget(self.plot_plugin)


class ControlPlaceholder(QWidget):
    def __init__(self, parent: Optional[QWidget] = None) -> None:
        super().__init__(parent)

        self.setLayout(QGridLayout(self))

        with importlib.resources.path(resources, "icon-black.svg") as path:
            icon = QSvgWidget(str(path), self)

        icon.setMaximumSize(60, 60)
        icon.renderer().setAspectRatioMode(QtCore.Qt.AspectRatioMode.KeepAspectRatio)
        effect = QGraphicsOpacityEffect(icon)
        effect.setOpacity(0.1)
        icon.setGraphicsEffect(effect)

        self.layout().addWidget(icon)


class PluginControlArea(QWidget):
    def __init__(self, app_model: AppModel, parent: QWidget) -> None:
        super().__init__(parent)

        self.app_model = app_model
        self.app_model.sig_load_plugin.connect(self._on_app_model_load_plugin)

        self.child_widget: QWidget = ControlPlaceholder()
        self.view_plugin: Optional[Any] = None

        self.setLayout(QVBoxLayout(self))
        self.layout().setContentsMargins(0, 0, 0, 0)
        self.layout().setSpacing(0)
        self.layout().addWidget(self.child_widget)

        if isinstance(app_model.plugin, PluginSpecBase):
            self._on_app_model_load_plugin(app_model.plugin)
        elif app_model.plugin is not None:
            raise RuntimeError(f"{type(app_model.plugin)} is not a PluginSpecBase.")

    def _on_app_model_load_plugin(self, plugin_spec: Optional[PluginSpecBase]) -> None:
        log.debug(
            f"{self.__class__.__name__} is going to replace its view_plugin"
            + f"({self.view_plugin.__class__.__name__}) in favour of {plugin_spec}"
        )
        self.view_plugin = None
        self.child_widget.deleteLater()

        if plugin_spec is None:
            self.child_widget = ControlPlaceholder()
        else:
            self.child_widget = plugin_spec.create_view_plugin(
                app_model=self.app_model,
            )

        self.layout().addWidget(self.child_widget)



# Copyright (c) Acconeer AB, 2023
# All rights reserved

from __future__ import annotations

import acconeer.exptool as et
from acconeer.exptool import a121
from acconeer.exptool.a121.algo.hand_motion import (
    AppMode,
    DetectionState,
    ModeHandler,
    ModeHandlerConfig,
)


sensor_id = 1


def main():
    client = a121.Client.open()

    config = ModeHandlerConfig()

    aggregator = ModeHandler(
        client=client,
        sensor_id=sensor_id,
        mode_handler_config=config,
    )

    aggregator.start()

    interrupt_handler = et.utils.ExampleInterruptHandler()
    print("Press Ctrl-C to end session")

    while not interrupt_handler.got_signal:
        result = aggregator.get_next()

        if result.app_mode == AppMode.PRESENCE:
            print("No presence detected")
        elif result.app_mode == AppMode.HANDMOTION:
            if result.example_app_result.detection_state == DetectionState.NO_DETECTION:
                print("No hand motion detected")
            elif result.example_app_result.detection_state == DetectionState.RETENTION:
                print("Retaining detection")
            elif result.example_app_result.detection_state == DetectionState.DETECTION:
                print("Hand motion detected")
            else:
                raise RuntimeError("Invalid detection state")

    print("Disconnecting...")


if __name__ == "__main__":
    main()



# Copyright (c) Acconeer AB, 2023
# All rights reserved

from __future__ import annotations

import warnings
from typing import Any, Dict, Optional, Tuple

import attrs
import h5py

from acconeer.exptool import a121
from acconeer.exptool.a121._h5_utils import _create_h5_string_dataset
from acconeer.exptool.a121.algo import Controller
from acconeer.exptool.a121.algo.distance import (
    Detector,
    DetectorConfig,
    DetectorContext,
    DetectorResult,
)

from ._processor import Processor, ProcessorConfig, ProcessorExtraResult, ProcessorLevelStatus


@attrs.mutable(kw_only=True)
class RefAppConfig(DetectorConfig):
    start_m: float = attrs.field(default=0.03)
    """Start of measurement range."""
    end_m: float = attrs.field(default=0.5)
    """End of measurement range."""
    median_filter_length: int = attrs.field(default=5)
    """Length of the median filter used to improve robustness of the result."""
    num_medians_to_average: int = attrs.field(default=1)
    """Number of medians averaged to obtain the final level."""

    @start_m.validator
    def _(self, _: Any, value: float) -> None:
        if value < Detector.MIN_DIST_M:
            raise ValueError(f"Cannot start measurements closer than {Detector.MIN_DIST_M}m")

    @end_m.validator
    def _(self, _: Any, value: float) -> None:
        if value > Detector.MAX_DIST_M:
            raise ValueError(f"Cannot measure further than {Detector.MAX_DIST_M}m")

    def to_detector_config(self) -> DetectorConfig:
        return DetectorConfig(
            start_m=self.start_m - 0.015,
            end_m=self.end_m * 1.05,
            max_step_length=self.max_step_length,
            max_profile=self.max_profile,
            close_range_leakage_cancellation=True,
            signal_quality=self.signal_quality,
            threshold_method=self.threshold_method,
            peaksorting_method=self.peaksorting_method,
            reflector_shape=self.reflector_shape,
            num_frames_in_recorded_threshold=self.num_frames_in_recorded_threshold,
            fixed_threshold_value=self.fixed_threshold_value,
            threshold_sensitivity=self.threshold_sensitivity,
            update_rate=self.update_rate,
        )


@attrs.frozen(kw_only=True)
class RefAppExtraResult:
    processor_extra_result: ProcessorExtraResult
    detector_result: Dict[int, DetectorResult]


RefAppContext = DetectorContext


@attrs.frozen(kw_only=True)
class RefAppResult:
    peak_detected: Optional[bool]
    """True if a peak (level) is detected, False if no peak is
    detected, or None if a result is not available."""
    peak_status: Optional[ProcessorLevelStatus]
    """Status assigned to the detected peak."""
    level: Optional[float]
    """Liquid level relative to the base of the tank."""
    extra_result: RefAppExtraResult
    """Extra result: Only used for the plots in the GUI."""


class RefApp(Controller[RefAppConfig, RefAppResult]):
    def __init__(
        self,
        *,
        client: a121.Client,
        sensor_id: int,
        config: RefAppConfig,
        context: Optional[RefAppContext] = None,
    ) -> None:
        super().__init__(client=client, config=config)
        self.sensor_id = sensor_id

        detector_config = self.config.to_detector_config()

        self._detector = Detector(
            client=self.client,
            sensor_ids=[self.sensor_id],
            detector_config=detector_config,
            context=context,
        )

        processor_config = ProcessorConfig(
            median_filter_length=self.config.median_filter_length,
            num_medians_to_average=self.config.num_medians_to_average,
            tank_range_start_m=self.config.start_m,
            tank_range_end_m=self.config.end_m,
        )

        self._processor = Processor(processor_config)

        self.started = False

    def calibrate(self) -> None:
        self._detector.calibrate_detector()

    def start(
        self, recorder: Optional[a121.Recorder] = None, algo_group: Optional[h5py.Group] = None
    ) -> None:
        if self.started:
            raise RuntimeError("Already started")

        if recorder is not None:
            if isinstance(recorder, a121.H5Recorder):
                algo_group = recorder.require_algo_group("tank_level")
                _record_algo_data(algo_group, self.sensor_id, self.config, self._detector.context)
            else:
                # Should never happen as we currently only have the H5Recorder
                warnings.warn("Will not save algo data")

        self._detector.start(recorder=recorder, _algo_group=algo_group)

        self.started = True

    def get_next(self) -> RefAppResult:
        if not self.started:
            raise RuntimeError("Not started")

        result = self._detector.get_next()

        processor_result = self._processor.process(result)

        ref_app_extra_result = RefAppExtraResult(
            processor_extra_result=processor_result.extra_result, detector_result=result
        )

        return RefAppResult(
            peak_detected=processor_result.peak_detected,
            peak_status=processor_result.peak_status,
            level=processor_result.filtered_level,
            extra_result=ref_app_extra_result,
        )

    def update_config(self, config: RefAppConfig) -> None:
        raise NotImplementedError

    def stop(self) -> Any:
        if not self.started:
            raise RuntimeError("Already stopped")

        recorder_result = self._detector.stop()

        self.started = False

        return recorder_result


def _record_algo_data(
    algo_group: h5py.Group, sensor_id: int, config: RefAppConfig, context: RefAppContext
) -> None:
    algo_group.create_dataset("sensor_id", data=sensor_id, track_times=False)

    _create_h5_string_dataset(algo_group, "config", config.to_json())

    context_group = algo_group.create_group("tank_level_context")
    context.to_h5(context_group)


def _load_algo_data(algo_group: h5py.Group) -> Tuple[int, RefAppConfig, RefAppContext]:
    sensor_id = int(algo_group["sensor_id"][()])
    config = RefAppConfig.from_json(algo_group["config"][()])

    context_group = algo_group["tank_level_context"]
    tank_level_context = DetectorContext.from_h5(context_group)

    return sensor_id, config, tank_level_context



# Copyright (c) Acconeer AB, 2022-2023
# All rights reserved

from __future__ import annotations

import abc
from enum import Enum
from typing import Any, Callable, Generic, Mapping, Optional, Sequence, Tuple, Type, TypeVar, cast

import attrs
import numpy as np

from PySide6 import QtCore, QtGui
from PySide6.QtWidgets import (
    QCheckBox,
    QComboBox,
    QDoubleSpinBox,
    QGridLayout,
    QHBoxLayout,
    QLabel,
    QLayout,
    QSlider,
    QSpinBox,
    QStackedWidget,
    QVBoxLayout,
    QWidget,
)

from acconeer.exptool.a121._core.entities import Criticality
from acconeer.exptool.app.new.ui.components.data_editor import DataEditor

from .common import MaybeIterable, as_sequence


def widget_wrap_layout(layout: QLayout) -> QWidget:
    dummy = QWidget()
    dummy.setLayout(layout)
    return dummy


PidgetHook = Callable[["Pidget", Mapping[str, "Pidget"]], None]
T = TypeVar("T")
EnumT = TypeVar("EnumT", bound=Enum)


def _hooks_converter(a: MaybeIterable[PidgetHook]) -> Sequence[PidgetHook]:
    return as_sequence(a)


@attrs.frozen(kw_only=True, slots=False)
class PidgetFactory(abc.ABC):
    name_label_text: str = attrs.field()
    name_label_tooltip: Optional[str] = None
    note_label_text: Optional[str] = None
    hooks: Sequence[PidgetHook] = attrs.field(factory=tuple, converter=_hooks_converter)

    @abc.abstractmethod
    def create(self, parent: QWidget) -> Pidget:
        ...

    @name_label_text.validator
    def check_for_whitespaces(self, attribute: Any, value: str) -> None:
        if value != value.strip():
            raise ValueError("Labels cannot start or end with a whitespace")

    @name_label_text.validator
    def check_label_text_format(self, attribute: Any, value: str) -> None:
        if len(value) > 0 and value[-1] != ":":
            raise ValueError("Labels have to end with ':'")


class Pidget(DataEditor[Any]):
    """Base class for a parameter-bound widget.

    A ``Pidget`` comes with a
    ``name`` label and an ``note`` label by default.
    """

    sig_update = QtCore.Signal(object)

    def __init__(self, factory: PidgetFactory, parent: QWidget) -> None:
        super().__init__(parent=parent)

        self.setLayout(QVBoxLayout(self))
        self.layout().setContentsMargins(0, 0, 0, 0)

        self._body_widget = QWidget(self)
        self.layout().addWidget(self._body_widget)

        self.__label_widget = QLabel(factory.name_label_text, parent=self._body_widget)
        if factory.name_label_tooltip is not None:
            self.__label_widget.setToolTip(factory.name_label_tooltip)

        self._body_layout = self._create_body_layout(self.__label_widget)
        self._body_widget.setLayout(self._body_layout)

        self.__note_widget = QLabel(parent=self)
        self.__note_widget.setWordWrap(True)
        self.__note_widget.setContentsMargins(5, 5, 5, 5)
        self.set_note_text(factory.note_label_text)
        self.layout().addWidget(self.__note_widget)

    def _create_body_layout(self, note_label_widget: QWidget) -> QLayout:
        """Called by Pidget.__init__"""

        layout = QHBoxLayout(self._body_widget)
        layout.setContentsMargins(0, 0, 0, 0)
        layout.addWidget(note_label_widget)
        return layout

    def set_note_text(
        self, message: Optional[str], criticality: Optional[Criticality] = None
    ) -> None:
        if not message:
            self.__note_widget.hide()
            return

        COLOR_MAP = {
            Criticality.ERROR: "#E6635A",
            Criticality.WARNING: "#FCC842",
            None: "white",
        }

        self.__note_widget.show()
        self.__note_widget.setText(message)
        self.__note_widget.setStyleSheet(
            f"background-color: {COLOR_MAP[criticality]}; color: white; font: bold italic;"
        )

    @abc.abstractmethod
    def set_data(self, value: Any) -> None:
        pass

    @abc.abstractmethod
    def get_data(self) -> Any:
        pass

    @property
    def is_ready(self) -> bool:
        return not self.__note_widget.isVisible()


@attrs.frozen(kw_only=True, slots=False)
class IntPidgetFactory(PidgetFactory):
    limits: Optional[Tuple[Optional[int], Optional[int]]] = None
    suffix: Optional[str] = None

    def create(self, parent: QWidget) -> IntPidget:
        return IntPidget(self, parent)


class IntPidget(Pidget):
    def __init__(self, factory: IntPidgetFactory, parent: QWidget) -> None:
        super().__init__(factory, parent)

        self._spin_box = _PidgetSpinBox(
            self._body_widget,
            limits=factory.limits,
            suffix=factory.suffix,
        )
        self._spin_box.valueChanged.connect(self.__on_changed)
        self._body_layout.addWidget(self._spin_box)

    def set_data(self, value: Any) -> None:
        assert isinstance(value, int)

        with QtCore.QSignalBlocker(self):
            self._spin_box.setValue(value)

    def get_data(self) -> int:
        return int(self._spin_box.value())

    def __on_changed(self) -> None:
        self.sig_update.emit(self._spin_box.value())


@attrs.frozen(kw_only=True, slots=False)
class FloatPidgetFactory(PidgetFactory):
    limits: Optional[Tuple[Optional[float], Optional[float]]] = None
    suffix: Optional[str] = None
    decimals: int = 1

    def create(self, parent: QWidget) -> FloatPidget:
        return FloatPidget(self, parent)


class FloatPidget(Pidget):
    def __init__(self, factory: FloatPidgetFactory, parent: QWidget) -> None:
        super().__init__(factory, parent)

        self._spin_box = _PidgetDoubleSpinBox(
            self._body_widget,
            limits=factory.limits,
            suffix=factory.suffix,
            decimals=factory.decimals,
        )
        self._spin_box.valueChanged.connect(self.__on_changed)
        self._body_layout.addWidget(self._spin_box)

    def set_data(self, value: Any) -> None:
        assert isinstance(value, (int, float))

        with QtCore.QSignalBlocker(self):
            self._spin_box.setValue(value)

    def get_data(self) -> float:
        return float(self._spin_box.value())

    def __on_changed(self) -> None:
        self.sig_update.emit(self._spin_box.value())


@attrs.frozen(kw_only=True, slots=False)
class FloatSliderPidgetFactory(FloatPidgetFactory):
    limits: Tuple[float, float]
    log_scale: bool = False
    show_limit_values: bool = True
    limit_texts: Optional[Tuple[Optional[str], Optional[str]]] = None

    def __attrs_post_init__(self) -> None:
        if self.log_scale:
            if self.limits[0] <= 0:
                raise ValueError("Lower limit must be > 0 when using log scale")

    def create(self, parent: QWidget) -> FloatSliderPidget:  # type: ignore[override]
        return FloatSliderPidget(self, parent)


class FloatSliderPidget(Pidget):
    def __init__(self, factory: FloatSliderPidgetFactory, parent: QWidget) -> None:
        super().__init__(factory, parent)

        self.__spin_box = _PidgetDoubleSpinBox(
            self._body_widget,
            limits=factory.limits,
            suffix=factory.suffix,
            decimals=factory.decimals,
        )
        self.__spin_box.valueChanged.connect(self.__on_spin_box_changed)
        self._body_layout.addWidget(self.__spin_box, 0, 1)  # type: ignore[call-arg]

        slider_widget = QWidget(self._body_widget)
        self._body_layout.addWidget(slider_widget, 1, 0, 1, -1)  # type: ignore[call-arg]
        slider_widget.setLayout(QHBoxLayout(slider_widget))
        slider_widget.layout().setContentsMargins(11, 6, 11, 0)

        lower_limit, upper_limit = factory.limits
        if factory.show_limit_values:
            slider_widget.layout().addWidget(QLabel(str(lower_limit), slider_widget))

        self._slider = _PidgetFloatSlider(
            slider_widget,
            limits=factory.limits,
            decimals=factory.decimals,
            log_scale=factory.log_scale,
        )
        self._slider.wrapped_value_changed.connect(self.__on_slider_changed)
        slider_widget.layout().addWidget(self._slider)

        if factory.show_limit_values:
            slider_widget.layout().addWidget(QLabel(str(upper_limit), slider_widget))

        if factory.limit_texts is not None:
            label_text_widget = QWidget(self._body_widget)
            self._body_layout.addWidget(label_text_widget, 2, 0, 1, -1)  # type: ignore[call-arg]
            label_text_widget.setLayout(QHBoxLayout(label_text_widget))
            label_text_widget.layout().setContentsMargins(11, 0, 11, 0)
            label_text_widget.layout().setSpacing(0)

            left, right = factory.limit_texts

            if left is not None:
                label = QLabel(left, label_text_widget)
                label_text_widget.layout().addWidget(label)

            label_text_widget.layout().addStretch(1)  # type: ignore[attr-defined]

            if right is not None:
                label = QLabel(right, label_text_widget)
                label_text_widget.layout().addWidget(label)

    def _create_body_layout(self, note_label_widget: QWidget) -> QLayout:
        """Called by Pidget.__init__"""

        layout = QGridLayout(self._body_widget)
        layout.setContentsMargins(0, 0, 0, 6)
        layout.setSpacing(0)
        layout.addWidget(note_label_widget, 0, 0)
        return layout

    def set_data(self, value: Any) -> None:
        assert isinstance(value, (int, float))

        with QtCore.QSignalBlocker(self):
            self.__spin_box.setValue(value)
            self._slider.wrapped_set_value(value)

    def get_data(self) -> float:
        return float(self.__spin_box.value())

    def __on_spin_box_changed(self, value: float) -> None:
        with QtCore.QSignalBlocker(self):
            self._slider.wrapped_set_value(value)

        self.sig_update.emit(value)

    def __on_slider_changed(self, value: float) -> None:
        with QtCore.QSignalBlocker(self):
            self.__spin_box.setValue(value)

        self.sig_update.emit(value)


@attrs.frozen(kw_only=True, slots=False)
class OptionalPidgetFactory(PidgetFactory):
    checkbox_label_text: Optional[str] = None


class OptionalPidget(Pidget):
    """Optional parameter, not optional widget"""

    def __init__(self, factory: OptionalPidgetFactory, parent: QWidget) -> None:
        super().__init__(factory, parent)

        self._optional_widget = QWidget(self._body_widget)
        self._body_layout.addWidget(self._optional_widget)

        self._none_checkbox = QCheckBox(self._optional_widget)
        if factory.checkbox_label_text:
            self._none_checkbox.setText(factory.checkbox_label_text)

        self._optional_layout = self._create_optional_layout(self._none_checkbox)

    def _create_optional_layout(self, none_checkbox: QWidget) -> QLayout:
        """Called by OptionalPidget.__init__"""

        layout = QHBoxLayout(self._optional_widget)
        layout.setContentsMargins(0, 0, 0, 0)
        layout.addStretch(1)
        layout.addWidget(none_checkbox)
        return layout

    def set_data(self, value: Any) -> None:
        with QtCore.QSignalBlocker(self):
            self._none_checkbox.setChecked(value is not None)

    @abc.abstractmethod
    def get_data(self) -> Optional[Any]:
        pass


@attrs.frozen(kw_only=True, slots=False)
class OptionalIntPidgetFactory(OptionalPidgetFactory, IntPidgetFactory):
    init_set_value: Optional[int] = None

    def create(self, parent: QWidget) -> OptionalIntPidget:  # type: ignore[override]
        return OptionalIntPidget(self, parent)


class OptionalIntPidget(OptionalPidget):
    def __init__(self, factory: OptionalIntPidgetFactory, parent: QWidget) -> None:
        super().__init__(factory, parent)

        self._spin_box = _PidgetSpinBox(
            self._optional_widget,
            limits=factory.limits,
            suffix=factory.suffix,
            init_set_value=factory.init_set_value,
        )
        self._optional_layout.addWidget(self._spin_box)

        self._none_checkbox.stateChanged.connect(self.__on_changed)
        self._spin_box.valueChanged.connect(self.__on_changed)

    def __on_changed(self) -> None:
        checked = self._none_checkbox.isChecked()

        with QtCore.QSignalBlocker(self):
            self._spin_box.setEnabled(checked)

        value = self._spin_box.value() if checked else None
        self.sig_update.emit(value)

    def set_data(self, value: Any) -> None:
        super().set_data(value)

        with QtCore.QSignalBlocker(self):
            if value is None:
                self._spin_box.setEnabled(False)
            else:
                self._spin_box.setValue(value)
                self._spin_box.setEnabled(True)

    def get_data(self) -> Optional[int]:
        if not self._none_checkbox.isChecked():
            return None
        else:
            return int(self._spin_box.value())


@attrs.frozen(kw_only=True, slots=False)
class OptionalFloatPidgetFactory(OptionalPidgetFactory, FloatPidgetFactory):
    init_set_value: Optional[float] = None
    """
    The initial value of the spinbox (ignoring the None-ness of the pidget)
    """

    placeholder_text: Optional[str] = None
    """
    Text that will be displayed when the checkbox is unchecked (and data is None).
    placeholder_text = None disables the placeholder text entierly
    """

    def create(self, parent: QWidget) -> OptionalFloatPidget:  # type: ignore[override]
        return OptionalFloatPidget(self, parent)


class OptionalFloatPidget(OptionalPidget):
    _SPINBOX_INDEX = 0
    _DUMMY_INDEX = 1

    def __init__(self, factory: OptionalFloatPidgetFactory, parent: QWidget) -> None:
        super().__init__(factory, parent)

        self._container = QStackedWidget()
        self._spin_box = _PidgetDoubleSpinBox(
            self._container,
            decimals=factory.decimals,
            limits=factory.limits,
            suffix=factory.suffix,
            init_set_value=factory.init_set_value,
        )

        self._container.addWidget(self._spin_box)

        if factory.placeholder_text is None:
            self._dummy_spin_box = None
        else:
            self._dummy_spin_box = _PidgetDoubleSpinBox.placeholder_dummy(
                self._container, factory.placeholder_text
            )
            self._dummy_spin_box.setEnabled(False)
            self._container.addWidget(self._dummy_spin_box)

        self._optional_layout.addWidget(self._container)
        self._none_checkbox.stateChanged.connect(self.__on_changed)
        self._spin_box.valueChanged.connect(self.__on_changed)

    def __on_changed(self) -> None:
        checked = self._none_checkbox.isChecked()

        with QtCore.QSignalBlocker(self):
            if self._dummy_spin_box is None:
                self._spin_box.setEnabled(checked)
            else:
                self._container.setCurrentIndex(
                    self._SPINBOX_INDEX if checked else self._DUMMY_INDEX
                )

        value = self._spin_box.value() if checked else None
        self.sig_update.emit(value)

    def set_data(self, value: Any) -> None:
        super().set_data(value)

        if value is not None:
            with QtCore.QSignalBlocker(self):
                self._spin_box.setValue(value)

        if self._dummy_spin_box is None:
            self._spin_box.setEnabled(value is not None)
        else:
            self._container.setCurrentIndex(
                self._SPINBOX_INDEX if value is not None else self._DUMMY_INDEX
            )

    def get_data(self) -> Optional[float]:
        if not self._none_checkbox.isChecked():
            return None
        else:
            return float(self._spin_box.value())


@attrs.frozen(kw_only=True, slots=False)
class CheckboxPidgetFactory(PidgetFactory):
    name_label_text: str = attrs.field()

    def create(self, parent: QWidget) -> CheckboxPidget:
        return CheckboxPidget(self, parent)

    @name_label_text.validator
    def check_for_whitespaces(self, attribute: Any, value: str) -> None:
        if value != value.strip():
            raise ValueError("Labels cannot start or end with a whitespace")


class CheckboxPidget(Pidget):
    def __init__(self, factory: CheckboxPidgetFactory, parent: QWidget) -> None:
        super().__init__(factory, parent)

        assert isinstance(self._body_layout, QGridLayout)

        self._checkbox = QCheckBox(self._body_widget)
        self._checkbox.clicked.connect(self.__on_checkbox_click)
        self._body_layout.addWidget(self._checkbox, 0, 0)
        self._body_layout.setColumnStretch(0, 0)
        self._body_layout.setColumnStretch(1, 1)

    def _create_body_layout(self, note_label_widget: QWidget) -> QLayout:
        """Called by Pidget.__init__"""

        layout = QGridLayout(self._body_widget)
        layout.setContentsMargins(0, 0, 0, 0)
        layout.setSpacing(0)
        layout.addWidget(note_label_widget, 0, 1)
        return layout

    def __on_checkbox_click(self, checked: bool) -> None:
        self.sig_update.emit(checked)

    def set_data(self, param: Any) -> None:
        with QtCore.QSignalBlocker(self):
            self._checkbox.setChecked(bool(param))

    def get_data(self) -> bool:
        return bool(self._checkbox.isChecked())


@attrs.frozen(kw_only=True, slots=False)
class ComboboxPidgetFactory(PidgetFactory, Generic[T]):
    items: list[tuple[str, T]]

    def create(self, parent: QWidget) -> ComboboxPidget[T]:
        return ComboboxPidget(self, parent)


class ComboboxPidget(Pidget, Generic[T]):
    def __init__(self, factory: ComboboxPidgetFactory[T], parent: QWidget) -> None:
        super().__init__(factory, parent)

        self._combobox = PidgetComboBox(self._body_widget)
        self._body_layout.addWidget(self._combobox)

        for displayed_text, user_data in factory.items:
            self._combobox.addItem(displayed_text, user_data)

        self._combobox.currentIndexChanged.connect(self._emit_data_of_combobox_item)

    def _emit_data_of_combobox_item(self, index: int) -> None:
        data = self._combobox.itemData(index)
        self.sig_update.emit(data)

    def set_data(self, param: Any) -> None:
        with QtCore.QSignalBlocker(self):
            index = self._combobox.findData(param)
            if index == -1:
                raise ValueError(f"Data item {param} could not be found in {self}.")
            self._combobox.setCurrentIndex(index)

    def get_data(self) -> T:
        return cast(T, self._combobox.currentData())


@attrs.frozen(kw_only=True, slots=False)
class SensorIdPidgetFactory(ComboboxPidgetFactory[int]):
    name_label_text: str = attrs.field(default="Sensor:")
    name_label_tooltip: str = attrs.field(default="The sensor to use in session")

    def create(self, parent: QWidget) -> SensorIdPidget:
        return SensorIdPidget(self, parent)


class SensorIdPidget(ComboboxPidget[int]):
    def __init__(self, factory: SensorIdPidgetFactory, parent: QWidget) -> None:
        super().__init__(factory, parent)
        self._sensor_id = 1
        self.set_data(self._sensor_id)

    def _emit_data_of_combobox_item(self, index: int) -> None:
        data = self._combobox.itemData(index)

        if data is None:
            return

        self._sensor_id = data
        self.sig_update.emit(data)

    def set_selectable_sensors(self, sensor_list: list[int]) -> None:
        with QtCore.QSignalBlocker(self):
            self._combobox.clear()

            for sensor_id in sensor_list:
                self._combobox.addItem(str(sensor_id), sensor_id)

            self.set_data(self._sensor_id)

    def set_data(self, sensor_id: int) -> None:
        self._sensor_id = sensor_id

        try:
            super().set_data(sensor_id)
        except ValueError:
            with QtCore.QSignalBlocker(self):
                self._combobox.setCurrentIndex(-1)

            self._combobox.setPlaceholderText(f"{self._sensor_id} (unavailable)")

    def get_data(self) -> int:
        return self._sensor_id


@attrs.frozen(kw_only=True, slots=False)
class EnumPidgetFactory(ComboboxPidgetFactory[EnumT]):
    enum_type: Type[EnumT] = attrs.field()
    label_mapping: dict[EnumT, str] = attrs.field()

    items: list[tuple[str, EnumT]] = attrs.field(init=False)

    def __attrs_post_init__(self) -> None:
        if self.label_mapping.keys() != set(self.enum_type):
            raise ValueError("label_mapping does not match enum_type")

        items = [(v, k) for k, v in self.label_mapping.items()]

        # The instance is immutable at this point, which is circumvented by the next row. See:
        # - https://www.attrs.org/en/stable/api.html#attr.ib
        # - https://github.com/python-attrs/attrs/issues/120
        # - https://github.com/python-attrs/attrs/issues/147

        object.__setattr__(self, "items", items)

    def create(self, parent: QWidget) -> EnumPidget[EnumT]:
        return EnumPidget[EnumT](self, parent)


class EnumPidget(ComboboxPidget[EnumT]):
    def __init__(self, factory: EnumPidgetFactory[EnumT], parent: QWidget) -> None:
        super().__init__(factory, parent)


@attrs.frozen(kw_only=True, slots=False)
class OptionalEnumPidgetFactory(OptionalPidgetFactory, EnumPidgetFactory[EnumT]):
    def create(self, parent: QWidget) -> OptionalEnumPidget:  # type: ignore[override]
        return OptionalEnumPidget(self, parent)


class OptionalEnumPidget(OptionalPidget):
    def __init__(self, factory: OptionalEnumPidgetFactory[EnumT], parent: QWidget) -> None:
        super().__init__(factory, parent)

        self._combobox = PidgetComboBox(self._body_widget)
        self._body_layout.addWidget(self._combobox)

        for displayed_text, user_data in factory.items:
            self._combobox.addItem(displayed_text, user_data)

        self._none_checkbox.stateChanged.connect(self.__emit_data_of_combobox_or_none)
        self._none_checkbox.stateChanged.connect(self.__enable_combobox_if_checked)
        self._combobox.currentIndexChanged.connect(self.__emit_data_of_combobox_or_none)

    def __emit_data_of_combobox_or_none(self) -> None:
        if self._none_checkbox.isChecked():
            data = self._combobox.currentData()
            self.sig_update.emit(data)
        else:
            self.sig_update.emit(None)

    def __enable_combobox_if_checked(self) -> None:
        self._combobox.setEnabled(self._none_checkbox.isChecked())

    def set_data(self, value: Any) -> None:
        super().set_data(value)
        if value is not None:
            self.set_enum_parameter(value)
            self._combobox.setEnabled(True)
        else:
            self._combobox.setEnabled(False)

    def get_data(self) -> Optional[Enum]:
        if not self._none_checkbox.isChecked():
            return None
        else:
            return cast(Enum, self._combobox.currentData())

    def set_enum_parameter(self, param: Any) -> None:
        with QtCore.QSignalBlocker(self):
            index = self._combobox.findData(param)
            if index == -1:
                raise ValueError(f"Data item {param} could not be found in {self}.")
            self._combobox.setCurrentIndex(index)


WIDGET_WIDTH = 125


class PidgetComboBox(QComboBox):
    def __init__(self, parent: QWidget) -> None:
        super().__init__(parent)

        self.setFocusPolicy(QtCore.Qt.FocusPolicy.StrongFocus)
        self.setFixedWidth(WIDGET_WIDTH)

    def wheelEvent(self, event: QtGui.QWheelEvent) -> None:
        if self.hasFocus():
            super().wheelEvent(event)
        else:
            event.ignore()


class _PidgetSpinBox(QSpinBox):
    def __init__(
        self,
        parent: QWidget,
        *,
        limits: Optional[Tuple[Optional[int], Optional[int]]] = None,
        init_set_value: Optional[int] = None,
        suffix: Optional[str] = None,
    ) -> None:
        super().__init__(parent)

        self.setKeyboardTracking(False)
        self.setFocusPolicy(QtCore.Qt.FocusPolicy.StrongFocus)
        self.setAlignment(QtCore.Qt.AlignmentFlag.AlignRight)
        self.setFixedWidth(WIDGET_WIDTH)

        self.setRange(*_convert_int_limits_to_qt_range(limits))

        if suffix:
            self.setSuffix(f" {suffix}")

        if init_set_value is not None:
            self.setValue(init_set_value)

    def wheelEvent(self, event: QtGui.QWheelEvent) -> None:
        if self.hasFocus():
            super().wheelEvent(event)
        else:
            event.ignore()


class _PidgetDoubleSpinBox(QDoubleSpinBox):
    def __init__(
        self,
        parent: QWidget,
        *,
        limits: Optional[Tuple[Optional[float], Optional[float]]] = None,
        init_set_value: Optional[float] = None,
        decimals: int = 1,
        suffix: Optional[str] = None,
    ) -> None:
        super().__init__(parent)

        self.setKeyboardTracking(False)
        self.setFocusPolicy(QtCore.Qt.FocusPolicy.StrongFocus)
        self.setAlignment(QtCore.Qt.AlignmentFlag.AlignRight)
        self.setFixedWidth(WIDGET_WIDTH)

        self.setRange(*_convert_float_limits_to_qt_range(limits))
        self.setDecimals(decimals)
        self.setSingleStep(10 ** (-decimals))

        if suffix:
            self.setSuffix(f" {suffix}")

        if init_set_value is not None:
            self.setValue(init_set_value)

    @classmethod
    def placeholder_dummy(cls, parent: QWidget, placeholder_text: str) -> _PidgetDoubleSpinBox:
        """
        Create a spinbox that only displays a placeholder text
        """
        spin_box = _PidgetDoubleSpinBox(parent)
        spin_box.setSpecialValueText(placeholder_text)
        spin_box.setRange(0, 1)
        spin_box.setValue(0)
        spin_box.setReadOnly(True)
        return spin_box

    def wheelEvent(self, event: QtGui.QWheelEvent) -> None:
        if self.hasFocus():
            super().wheelEvent(event)
        else:
            event.ignore()


class _PidgetFloatSlider(QSlider):
    NUM_STEPS = 1000

    wrapped_value_changed = QtCore.Signal(float)

    def __init__(
        self,
        parent: QWidget,
        *,
        limits: Tuple[float, float],
        decimals: int,
        log_scale: bool,
    ) -> None:
        super().__init__(QtCore.Qt.Orientation.Horizontal, parent)

        self.limits = limits
        self.decimals = decimals
        self.log_scale = log_scale

        self.setRange(0, self.NUM_STEPS)
        self.setFocusPolicy(QtCore.Qt.FocusPolicy.StrongFocus)

        self.valueChanged.connect(self.__on_value_changed)

    def __on_value_changed(self, value: int) -> None:
        self.wrapped_value_changed.emit(self.__from_slider_scale(value))

    def wrapped_set_value(self, value: float) -> None:
        self.setValue(self.__to_slider_scale(value))

    def wheelEvent(self, event: QtGui.QWheelEvent) -> None:
        if self.hasFocus():
            super().wheelEvent(event)
        else:
            event.ignore()

    def __to_slider_scale(self, x: float) -> int:
        lower, upper = self.limits

        if self.log_scale:
            lower = np.log(lower)
            upper = np.log(upper)
            x = np.log(x)

        y = (x - lower) / (upper - lower) * self.NUM_STEPS
        return int(round(y))

    def __from_slider_scale(self, y: int) -> float:
        lower, upper = self.limits

        if self.log_scale:
            lower = np.log(lower)
            upper = np.log(upper)

        x = y / self.NUM_STEPS * (upper - lower) + lower

        if self.log_scale:
            x = np.exp(x)

        return round(x, self.decimals)


def _convert_int_limits_to_qt_range(
    limits: Optional[Tuple[Optional[int], Optional[int]]]
) -> Tuple[int, int]:
    if limits is None:
        limits = (None, None)

    lower, upper = limits

    if lower is None:
        lower = int(-1e9)

    if upper is None:
        upper = int(1e9)

    return (lower, upper)


def _convert_float_limits_to_qt_range(
    limits: Optional[Tuple[Optional[float], Optional[float]]]
) -> Tuple[float, float]:
    if limits is None:
        limits = (None, None)

    lower, upper = limits

    if lower is None:
        lower = -1e9

    if upper is None:
        upper = 1e9

    return (lower, upper)



# Copyright (c) Acconeer AB, 2022-2024
# All rights reserved

from __future__ import annotations

from .app_model import PluginSpec


def load_default_plugins() -> list[PluginSpec]:
    from acconeer.exptool.a121.algo.bilateration._plugin import BILATERATION_PLUGIN
    from acconeer.exptool.a121.algo.breathing._ref_app_plugin import BREATHING_PLUGIN
    from acconeer.exptool.a121.algo.distance._detector_plugin import DISTANCE_DETECTOR_PLUGIN
    from acconeer.exptool.a121.algo.hand_motion._example_app_plugin import HAND_MOTION_PLUGIN
    from acconeer.exptool.a121.algo.obstacle._detector_plugin import OBSTACLE_DETECTOR_PLUGIN
    from acconeer.exptool.a121.algo.obstacle_bilateration._plugin import (
        OBSTACLE_BILATERATION_PLUGIN,
    )
    from acconeer.exptool.a121.algo.parking._ref_app_plugin import PARKING_PLUGIN
    from acconeer.exptool.a121.algo.phase_tracking._plugin import PHASE_TRACKING_PLUGIN
    from acconeer.exptool.a121.algo.presence._detector_plugin import PRESENCE_DETECTOR_PLUGIN
    from acconeer.exptool.a121.algo.smart_presence._ref_app_plugin import SMART_PRESENCE_PLUGIN
    from acconeer.exptool.a121.algo.sparse_iq._plugin import SPARSE_IQ_PLUGIN
    from acconeer.exptool.a121.algo.speed._detector_plugin import SPEED_DETECTOR_PLUGIN
    from acconeer.exptool.a121.algo.surface_velocity._example_app_plugin import (
        SURFACE_VELOCITY_PLUGIN,
    )
    from acconeer.exptool.a121.algo.tank_level._plugin import TANK_LEVEL_PLUGIN
    from acconeer.exptool.a121.algo.touchless_button._plugin import TOUCHLESS_BUTTON_PLUGIN
    from acconeer.exptool.a121.algo.vibration._plugin import VIBRATION_PLUGIN

    # Please keep in lexicographical order
    return [
        BILATERATION_PLUGIN,
        BREATHING_PLUGIN,
        DISTANCE_DETECTOR_PLUGIN,
        HAND_MOTION_PLUGIN,
        OBSTACLE_BILATERATION_PLUGIN,
        OBSTACLE_DETECTOR_PLUGIN,
        PARKING_PLUGIN,
        PHASE_TRACKING_PLUGIN,
        PRESENCE_DETECTOR_PLUGIN,
        SMART_PRESENCE_PLUGIN,
        SPARSE_IQ_PLUGIN,
        SURFACE_VELOCITY_PLUGIN,
        TANK_LEVEL_PLUGIN,
        TOUCHLESS_BUTTON_PLUGIN,
        VIBRATION_PLUGIN,
        SPEED_DETECTOR_PLUGIN,
    ]



# Copyright (c) Acconeer AB, 2022-2024
# All rights reserved

from __future__ import annotations

import logging

import attrs
import numpy as np

# Added here to force pyqtgraph to choose PySide
import PySide6  # noqa: F401

import pyqtgraph as pg

import acconeer.exptool as et
from acconeer.exptool import a121
from acconeer.exptool.a121.algo.distance import (
    PeakSortingMethod,
    ReflectorShape,
    ThresholdMethod,
)
from acconeer.exptool.a121.algo.distance._processors import (
    DEFAULT_FIXED_AMPLITUDE_THRESHOLD_VALUE,
    DEFAULT_FIXED_STRENGTH_THRESHOLD_VALUE,
)
from acconeer.exptool.a121.algo.tank_level import RefApp, RefAppResult
from acconeer.exptool.a121.algo.tank_level._processor import ProcessorLevelStatus
from acconeer.exptool.a121.algo.tank_level._ref_app import RefAppConfig, RefAppContext


log = logging.getLogger(__name__)


TIME_HISTORY_S = 30


@attrs.mutable(kw_only=True)
class SharedState:
    sensor_id: int = attrs.field(default=1)
    config: RefAppConfig = attrs.field(factory=RefAppConfig)
    context: RefAppContext = attrs.field(factory=RefAppContext)


def main():
    args = a121.ExampleArgumentParser().parse_args()
    et.utils.config_logging(args)

    # Setup the configurations
    # Detailed at https://docs.acconeer.com/en/latest/exploration_tool/algo/a121/ref_apps/tank_level.html

    # Sensor selections
    sensor = 1

    # Tank level configurations
    ref_app_config = RefAppConfig(
        start_m=0.03,
        end_m=0.5,
        max_step_length=2,
        max_profile=a121.Profile.PROFILE_2,
        close_range_leakage_cancellation=True,
        signal_quality=20,
        update_rate=None,
        median_filter_length=5,
        num_medians_to_average=5,
        threshold_method=ThresholdMethod.CFAR,
        reflector_shape=ReflectorShape.PLANAR,
        peaksorting_method=PeakSortingMethod.CLOSEST,
        num_frames_in_recorded_threshold=50,
        fixed_threshold_value=DEFAULT_FIXED_AMPLITUDE_THRESHOLD_VALUE,  # float
        fixed_strength_threshold_value=DEFAULT_FIXED_STRENGTH_THRESHOLD_VALUE,  # float
        threshold_sensitivity=0.0,  # float
    )

    # End setup configurations

    # Preparation for client
    client = a121.Client.open(**a121.get_client_args(args))

    # Preparation for reference application processor
    ref_app = RefApp(client=client, sensor_id=sensor, config=ref_app_config)
    ref_app.calibrate()
    ref_app.start()

    pg_updater = PGUpdater(
        config=ref_app_config, num_curves=len(ref_app._detector.processor_specs)
    )
    pg_process = et.PGProcess(pg_updater, max_freq=60)
    pg_process.start()

    interrupt_handler = et.utils.ExampleInterruptHandler()
    print("Press Ctrl-C to end session")

    while not interrupt_handler.got_signal:
        processed_data = ref_app.get_next()
        try:
            pg_process.put_data(processed_data)
        except et.PGProccessDiedException:
            break

    ref_app.stop()
    client.close()
    print("Disconnecting...")


class PGUpdater:
    STATUS_MSG_MAP = {
        ProcessorLevelStatus.IN_RANGE: "In range",
        ProcessorLevelStatus.NO_DETECTION: "Not available",
        ProcessorLevelStatus.OVERFLOW: "Warning: Overflow",
        ProcessorLevelStatus.OUT_OF_RANGE: "Out of range",
    }

    def __init__(
        self,
        config: RefAppConfig,
        num_curves: int,
    ) -> None:
        self.num_curves = num_curves
        self.start_m = config.start_m
        self.end_m = config.end_m

    def setup(self, win):
        # Sweep plot
        self.sweep_plot = win.addPlot(row=1, col=0, colspan=3)
        self.sweep_plot.setMenuEnabled(False)
        self.sweep_plot.showGrid(x=True, y=True)
        self.sweep_plot.addLegend()
        self.sweep_plot.setLabel("left", "Amplitude")
        self.sweep_plot.setLabel("bottom", "Distance (m)")
        self.sweep_plot.addItem(pg.PlotDataItem())

        self.vertical_line_start = pg.InfiniteLine(
            pen=et.utils.pg_pen_cycler(2),
            label="Tank start",
            labelOpts={
                "position": 0.5,
                "color": (0, 100, 0),
                "fill": (200, 200, 200, 50),
                "movable": True,
            },
        )
        self.sweep_plot.addItem(self.vertical_line_start)
        self.vertical_line_end = pg.InfiniteLine(
            pen=et.utils.pg_pen_cycler(2),
            label="Tank end",
            labelOpts={
                "position": 0.5,
                "color": (0, 100, 0),
                "fill": (200, 200, 200, 50),
                "movable": True,
            },
        )
        self.sweep_plot.addItem(self.vertical_line_end)

        pen = et.utils.pg_pen_cycler(0)
        brush = et.utils.pg_brush_cycler(0)
        symbol_kw = dict(symbol="o", symbolSize=1, symbolBrush=brush, symbolPen="k")
        feat_kw = dict(pen=pen, **symbol_kw)
        self.sweep_curves = [self.sweep_plot.plot(**feat_kw) for _ in range(self.num_curves)]

        pen = et.utils.pg_pen_cycler(1)
        brush = et.utils.pg_brush_cycler(1)
        symbol_kw = dict(symbol="o", symbolSize=1, symbolBrush=brush, symbolPen="k")
        feat_kw = dict(pen=pen, **symbol_kw)
        self.threshold_curves = [self.sweep_plot.plot(**feat_kw) for _ in range(self.num_curves)]

        sweep_plot_legend = pg.LegendItem(offset=(0.0, 0.5))
        sweep_plot_legend.setParentItem(self.sweep_plot)
        sweep_plot_legend.addItem(self.sweep_curves[0], "Sweep")
        sweep_plot_legend.addItem(self.threshold_curves[0], "Threshold")

        # Level history plot
        self.level_history_plot = win.addPlot(row=0, col=1, colspan=2)
        self.level_history_plot.setMenuEnabled(False)
        self.level_history_plot.showGrid(x=True, y=True)
        self.level_history_plot.addLegend()
        self.level_history_plot.setLabel("left", "Estimated level (cm)")
        self.level_history_plot.setLabel("bottom", "Time (s)")
        self.level_history_plot.addItem(pg.PlotDataItem())

        pen = et.utils.pg_pen_cycler(0)
        brush = et.utils.pg_brush_cycler(0)
        symbol_kw = dict(symbol="o", symbolSize=5, symbolBrush=brush, symbolPen="k")
        feat_kw = dict(pen=pen, **symbol_kw)
        self.level_history_curve = self.level_history_plot.plot(**feat_kw)

        self.sweep_smooth_max = et.utils.SmoothMax()
        self.distance_hist_smooth_lim = et.utils.SmoothLimits()

        # text items
        self.level_html_format = (
            '<div style="text-align: center">'
            '<span style="color: #FFFFFF;font-size:12pt;">'
            "{}</span></div>"
        )

        self.level_text_item = pg.TextItem(
            fill=pg.mkColor(0x1F, 0x77, 0xB4, 180),
            anchor=(0.5, 0),
        )
        self.level_history_plot.addItem(self.level_text_item)
        self.level_text_item.hide()

    def update(
        self,
        result: RefAppResult,
    ) -> None:
        # Get the first element as the plugin only supports single sensor operation.
        (detector_result,) = list(result.extra_result.detector_result.values())
        assert detector_result.distances is not None

        time_and_level_dict = (
            result.extra_result.processor_extra_result.level_and_time_for_plotting
        )

        # update sweep plot
        max_val_in_plot = 0
        for idx, processor_result in enumerate(detector_result.processor_results):
            assert processor_result.extra_result.used_threshold is not None
            assert processor_result.extra_result.distances_m is not None
            assert processor_result.extra_result.abs_sweep is not None

            self.sweep_curves[idx].setData(
                processor_result.extra_result.distances_m, processor_result.extra_result.abs_sweep
            )

            self.threshold_curves[idx].setData(
                processor_result.extra_result.distances_m,
                processor_result.extra_result.used_threshold,
            )

            max_val_in_subsweep = max(
                max(processor_result.extra_result.used_threshold),
                max(processor_result.extra_result.abs_sweep),
            )

            max_val_in_plot = max(max_val_in_plot, max_val_in_subsweep)

        self.sweep_plot.setYRange(0, self.sweep_smooth_max.update(max_val_in_plot))
        self.vertical_line_start.setValue(self.start_m)
        self.vertical_line_end.setValue(self.end_m)
        self.vertical_line_start.show()
        self.vertical_line_end.show()

        # update level history plot
        if any(~np.isnan(time_and_level_dict["level"])):
            self.level_history_curve.setData(
                time_and_level_dict["time"], time_and_level_dict["level"] * 100
            )
        self.level_history_plot.setXRange(-TIME_HISTORY_S + 1, 0)
        self.level_history_plot.setYRange(0, (self.end_m - self.start_m + 0.01) * 100)

        # update level plot
        if (
            result.level is not None
            and result.peak_detected is not None
            and result.peak_status is not None
        ):
            current_level = result.level
            peak_detected = result.peak_detected
            peak_status = result.peak_status
            level_text = self.STATUS_MSG_MAP[peak_status]
            if peak_detected:
                level_text = "Level: {:.1f} cm, {:.0f} %".format(
                    current_level * 100,
                    current_level / (self.end_m - self.start_m) * 100,
                )

                level_html = self.level_html_format.format(level_text)
                self.level_text_item.setHtml(level_html)

                x_pos = -(TIME_HISTORY_S) / 2
                y_max_cm = self.end_m * 100
                self.level_text_item.setPos(x_pos, 0.95 * y_max_cm)
                self.level_text_item.show()

            else:
                self.level_text_item.hide()


if __name__ == "__main__":
    main()
